/* Copyright (c) 2007-2010, Michal Bidlo (bidlom AT fit.vutbr.cz)
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <cmath>

using namespace std;

typedef unsigned int UINT;

#define DEBUG

///////////////////////////////////////////////////////////////
// sirky operandu nsobicek: 
// pro evoluci neni vhodne volit hodnoty vetsi nez 4
// pro testovani se doporucuje volit hodnoty v rozsahu 12 az 15
#define a_width 4
#define b_width 4
///////////////////////////////////////////////////////////////

#define _ID_V
#define _ID_HV
#define _AND
#define _HA_AV
#define _HA_AH
#define _FA_A
#define _HA_D
#define _FA_D   
#define _FA_AD
// #define _OR
// #define _XOR
// #define _NOT
// #define _ID_VH

// ---------------------- parametry genetickeho algoritmu ----------------------
const UINT pmut = 3;
const UINT pcross = 90;
const UINT unit = 100;
const UINT tour = 2;
const UINT popsize = 32;
const UINT avgfcalc = 500;

//////////////// struktury specificke pro ruzne druhy nasobicek ////////////////
// prostredi pro klasicke kombinacni nasobicky: a_width-2 "2"
const UINT env[] = { 0, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2 };
// prostredi pro nasobicky s uchovanim prenosu: a_width-1 "1"
// const UINT env[] = { 0, 1, 1, 1, 2 };
//
// delky programu pro klasicke kombinacni nasobicky
const UINT aprog_len[] = { 6, 12, 12 };
// delky programu pro nasobicky s uchovanim prenosu
// const UINT aprog_len[] = { 6, 10, 9 };
const UINT progs_len = 30;
//
// pocet stupnu obvodu pro klasicke kombinacni nasobicky
const UINT levels = b_width;
// pocet stupnu obvodu pro nasobicky s uchovanim prenosu
// const UINT levels = b_width + 1;
////////////////////////////////////////////////////////////////////////////////


const UINT bitw = a_width + b_width;

// rozmery pole pro vyvoj obvodu
const UINT rows = levels + 1;
const UINT cols = bitw + 1;

#ifdef DEBUG
	const UINT generations = 0;
#else
	const UINT generations = 1000000;
#endif

// maximalni ciselna konstanta pripustna pro instrukce
const UINT max_const = 2;

// promenne pouzivane pri developmentu
const UINT v0 = max_const + 1;
const UINT v1 = max_const + 2;
const UINT v2 = max_const + 3;
const UINT v3 = max_const + 4;

UINT var[4];

// parametry vyvojoveho systemu
const UINT p0 = max_const + 5;
const UINT p1 = max_const + 6;

// parametry jsou bitove sirky operandu
UINT par[2] = { a_width, b_width };

// symboly instrukci
enum INST
{
    SET,
    INC,
    DEC,
    REP,
    GEN,
    NOP,
    INSS
};

// symboly stavebnich bloku
enum BLOCK
{
#ifdef _ID_V
    ID_V,
#endif
#ifdef _ID_HV
    ID_HV,
#endif
#ifdef _AND
    AND,
#endif
#ifdef _HA_AV
    HA_AV,
#endif
#ifdef _HA_AH
    HA_AH,
#endif
#ifdef _FA_A
    FA_A,
#endif
#ifdef _HA_D
    HA_D,
#endif
#ifdef _FA_D   
    FA_D,   
#endif
#ifdef _FA_AD
    FA_AD,
#endif
#ifdef _OR
    OR,
#endif
#ifdef _XOR
    XOR,
#endif
#ifdef _NOT
    NOT,
#endif
#ifdef _ID_VH
    ID_VH,
#endif

    BLOCKS
};

// struktura elementu obvodu
struct Cell
{
    UINT block; // odkaz funkce
    UINT in1, in2;  // vstupy hradla and u scitacek
    UINT h_out; // hodnota praveho a
    UINT v_out; // spodniho vystuipu
    bool active;    // aktivni/neaktinvi
};

// struktura instrukce
struct Inst
{
    UINT opc;
    UINT arg1, arg2;
};

// struktura pro vyvoj obvodu
Cell circuit[rows][cols];
// aktualni pozice generatoru ve strukture
UINT lvl, pos;

// ----------------------------- implementace GA -------------------------------

UINT offset[sizeof(aprog_len) / sizeof(aprog_len[0])];

struct fit
{
    double pop_avg;
    double best;
    UINT gen;
};
fit pop_fit[generations / avgfcalc + 2];
UINT fit_ix;
double average_fit;

struct GA_chromosome	// datovy typ chromozomu
{
    Inst prog[progs_len];
	double fitness;
	bool evaluate;
};

double max_fit;
double best_ever;
GA_chromosome best;	// nejlepsi dosud nalezene reseni
    
UINT generation;
GA_chromosome *population;
GA_chromosome *next_population;
GA_chromosome pool1[popsize];
GA_chromosome pool2[popsize];

// struktury potrebne pro evaluaci obvodu
const UINT patterns[]={0xAAAAAAAA,0xCCCCCCCC,0xF0F0F0F0,0xFF00FF00,0xFFFF0000};

UINT sequences[bitw + 1];
UINT vectors[bitw + 1];

ostream & operator<<(ostream &os, const GA_chromosome &g);

UINT random(UINT low, UINT high)
{
    return (rand() % (high - low + 1)) + low;
}

void set_args(Inst &i)
{
        switch (i.opc)
        {
            case SET:
                i.arg1 = random(v0, v3);
                i.arg2 = random(0, p1);
                break;
            case INC:
            case DEC:
                i.arg1 = random(v0, v3);
                i.arg2 = random(0, max_const);
                break;
            case REP:
                i.arg1 = random(0, p1);
                i.arg2 = random(0, max_const);
                break;
            case GEN:
                i.arg1 = random(0, BLOCKS - 1);
                i.arg2 = 0;
                break;
            case NOP:
                i.arg1 = i.arg2 = 0;
                break;
            default: break;
        }
}

void initialize(GA_chromosome &genome)
{
    for (UINT i = 0; i < progs_len; i++)
    {
        genome.prog[i].opc = random(0, INSS - 1);
        set_args(genome.prog[i]);
    }
}

void set(UINT store, UINT source)
{
    if (source <= max_const)
        var[store - v0] = source;    // prirazeni ciselne konstanty
    else if (source < p0)
        var[store - v0] = var[source - v0]; // prirazeni hodnoty promenne
    else
        var[store - v0] = par[source - p0]; // prirazeni hodnoty parametru
}

void inc(UINT _var, UINT _val)
{
    var[_var - v0] += _val;
}

void dec(UINT _var, UINT _val)
{
    if (var[_var - v0] >= _val)
        var[_var - v0] -= _val;
}

void gen(UINT code)
{
    if (pos < cols)
    {
        circuit[lvl][pos].block = code;
        circuit[lvl][pos].active = true;
        circuit[lvl][pos].in1 = var[0];
        circuit[lvl][pos].in2 = var[1];
        pos++;
    }
}

void rep(Inst *prog, UINT ip, UINT limit)
{
    UINT cnt;
    if (prog[ip].arg1 < v0)
        cnt = prog[ip].arg1;        // pocet opakovani dan konstantou
    else if (prog[ip].arg1 < p0)
        cnt = var[prog[ip].arg1 - v0];   // pocet opak. dan hodnotou promenne
    else
        cnt = par[prog[ip].arg1 - p0];   // pocet opak. dan hodnotou parametru
    
    for (UINT i = 0; i < cnt; i++)	// arg1-krat opakovat arg2 instrukci za REP
    {
        for (UINT j = 1; ip + j < limit && j <= prog[ip].arg2; j++)
        {
            switch (prog[ip + j].opc)
            {
                case SET: set(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case INC: inc(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case DEC: dec(prog[ip + j].arg1, prog[ip + j].arg2); break;
                case GEN: gen(prog[ip + j].arg1); break;
                default: break;
            }
        }
    }
}

void run(Inst *prog, UINT limit)
{
    UINT ip = 0;
    while (ip < limit)
    {
        switch (prog[ip].opc)
        {
            case SET: set(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case INC: inc(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case DEC: dec(prog[ip].arg1, prog[ip].arg2); ip++; break;
            case REP: rep(prog, ip, limit); ip += prog[ip].arg2 + 1; break;
            case GEN: gen(prog[ip].arg1); ip++; break;
            default: ip++; break;
        }
    }
}

void develop(GA_chromosome &genome)
{
    // inicializace struktury - vsechny bunky neaktivni
    for (UINT i = 0; i < rows; i++)
    {
        for (UINT j = 0; j < cols; j++)
        {
            circuit[i][j].active = false;
            circuit[i][j].h_out = 0;
            circuit[i][j].v_out = 0;
        }
    }
    var[0] = 0; var[1] = 0; var[2] = 0; var[3] = 0;
    
    lvl = 1;
    for (UINT e = 0; e < levels; e++)
    {
        pos = 1;
        run(&genome.prog[offset[env[e]]], aprog_len[env[e]]);
        lvl++;
    }
}

double evaluate(UINT *input, UINT *result)
{
    UINT a, b, z;
    double corr = 0.0;
    
    for (UINT i = 0; i < 8 * sizeof(UINT); i++)
    {
        a = b = 0; z = 1;
        for (UINT j = 0; j < a_width; j++)
        {
            a += z * (input[j] & 1);
            input[j] >>= 1;
            z <<= 1;
        }
        z = 1;
        for (UINT j = 0; j < b_width; j++)
        {
            b += z * (input[a_width + j] & 1);
            input[a_width + j] >>= 1;
            z <<= 1;
        }
        z = a * b;
        for (UINT j = 0; j < bitw; j++)
        {
            if ((result[j] & 1) == (z & 1))
                corr++;
            result[j] >>= 1;
            z >>= 1;
        }
    }
    
    return corr;
}

double fitness(GA_chromosome &genome)
{
    double fit = 0.0;
	
    develop(genome);
    UINT test[bitw], inputs[bitw];
    UINT a = vectors[bitw] - 1, p;
    for (UINT i = 0; i < vectors[bitw]; i++)
    {
		p = a;
		// vytvoreni 32 testovacich sekvenci v podobe 1 unsigned integeru
		for (UINT j = 0; j < bitw; j++)
		{
		    if (j < 5) test[j] = patterns[j];
	    	else { test[j] = p & 1 ? 0xFFFFFFFF : 0; p >>= 1; }
		    inputs[j] = test[j];
		}
		a--;
        
		// simulace vypoctu obvodu s testovacimi sekvencemi na jeho vstupech
        UINT r;
        for (r = 1; r < rows && circuit[r][1].active; r++)
        {
            for (UINT c = 1; c < cols && circuit[r][c].active; c++)
            {
                switch (circuit[r][c].block)
                {
#ifdef _ID_V
                    case ID_V:
                        circuit[r][c].v_out = circuit[r - 1][c].v_out;
                        circuit[r][c].h_out = 0;
                        break;
#endif
#ifdef _ID_HV
                    case ID_HV:
                        circuit[r][c].v_out = circuit[r][c - 1].h_out;
                        circuit[r][c].h_out = 0;
                        break;
#endif
#ifdef _ID_VH
                    case ID_VH:
                        circuit[r][c].v_out = 0;
                        circuit[r][c].h_out = circuit[r - 1][c].v_out;
                        break;
#endif
#ifdef _AND
                    case AND:
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                circuit[r][c].v_out = test[circuit[r][c].in1] &
                                            test[a_width + circuit[r][c].in2];
                        else circuit[r][c].v_out = 0;
                        circuit[r][c].h_out = 0;
                        break;
#endif
#ifdef _OR
                    case OR:
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                circuit[r][c].v_out = test[circuit[r][c].in1] |
                                            test[a_width + circuit[r][c].in2];
                        else if (circuit[r][c].in1 < a_width)
                            circuit[r][c].v_out = test[circuit[r][c].in1];
                        else if (circuit[r][c].in2 < b_width)
                            circuit[r][c].v_out=test[a_width+circuit[r][c].in2];
                        else
                            circuit[r][c].v_out = 0;
                        circuit[r][c].h_out = 0;
                        break;
#endif
#ifdef _XOR
                    case XOR:
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                circuit[r][c].v_out = test[circuit[r][c].in1] ^
                                            test[a_width + circuit[r][c].in2];
                        else if (circuit[r][c].in1 < a_width)
                            circuit[r][c].v_out = test[circuit[r][c].in1];
                        else if (circuit[r][c].in2 < b_width)
                            circuit[r][c].v_out=test[a_width+circuit[r][c].in2];
                        else
                            circuit[r][c].v_out = 0;
                        circuit[r][c].h_out = 0;
                        break;
#endif
#ifdef _NOT
                    case NOT:
                        circuit[r][c].v_out = ~circuit[r - 1][c].v_out;
                        circuit[r][c].h_out = 0;
                    break;
#endif
#ifdef _HA_AV
                    case HA_AV: {
                        UINT _and;
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                _and = test[circuit[r][c].in1] &
                                       test[a_width + circuit[r][c].in2];
                        else _and = 0;
                        circuit[r][c].v_out = circuit[r - 1][c].v_out ^ _and;
                        circuit[r][c].h_out = circuit[r - 1][c].v_out & _and;
                        } break;
#endif
#ifdef _HA_AH
                    case HA_AH: {
                        UINT _and;
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                _and = test[circuit[r][c].in1] &
                                       test[a_width + circuit[r][c].in2];
                        else _and = 0;
                        circuit[r][c].v_out = circuit[r][c - 1].h_out ^ _and;
                        circuit[r][c].h_out = circuit[r][c - 1].h_out & _and;
                        } break;
#endif
#ifdef _HA_D
                    case HA_D:
                        circuit[r][c].v_out = circuit[r - 1][c - 1].h_out ^
                            circuit[r - 1][c].v_out;
                        circuit[r][c].h_out = circuit[r - 1][c - 1].h_out &
                            circuit[r - 1][c].v_out;
                        break;
#endif
#ifdef _FA_A
                    case FA_A: {
                        UINT _and;
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                            _and = test[circuit[r][c].in1] &
                                   test[a_width + circuit[r][c].in2];
                        else _and = 0;
                        circuit[r][c].v_out = circuit[r - 1][c].v_out ^ _and ^
                                              circuit[r][c - 1].h_out;
                        circuit[r][c].h_out = (circuit[r - 1][c].v_out & _and) |
                            (circuit[r - 1][c].v_out & circuit[r][c-1].h_out) |
                            (_and & circuit[r][c - 1].h_out);
                        } break;
#endif
#ifdef _FA_D
                    case FA_D:
                        circuit[r][c].v_out = circuit[r - 1][c - 1].h_out ^
                            circuit[r - 1][c].v_out ^ circuit[r][c - 1].h_out;
                        circuit[r][c].h_out =
                            (circuit[r-1][c-1].h_out & circuit[r-1][c].v_out) |
                            (circuit[r-1][c-1].h_out & circuit[r][c-1].h_out) |
                            (circuit[r-1][c].v_out & circuit[r][c-1].h_out);
                        break;
#endif
#ifdef _FA_AD
                    case FA_AD: {
                        UINT _and;
                        if (circuit[r][c].in1 < a_width &&
                            circuit[r][c].in2 < b_width)
                                _and = test[circuit[r][c].in1] &
                                       test[a_width + circuit[r][c].in2];
                        else _and = 0;
                        circuit[r][c].v_out = circuit[r - 1][c - 1].h_out ^
                            circuit[r - 1][c].v_out ^ _and;
                        circuit[r][c].h_out =
                            (circuit[r-1][c-1].h_out & circuit[r-1][c].v_out) |
                            (circuit[r-1][c-1].h_out & _and) |
                            (circuit[r - 1][c].v_out & _and);
                        } break;
#endif
                    default: break;
                }
            }
//            cout << endl;
        }
        // tady priradit vystupy obvodu zpet test signalum
        for (UINT t = 0; t < bitw; t++)
            test[t] = circuit[r - 1][t + 1].v_out;

		fit += evaluate(inputs, test);
	}
	
	return fit;
}

void crossover(GA_chromosome &parent1, GA_chromosome &parent2,
		GA_chromosome &child1, GA_chromosome &child2)
{
    child1 = parent1;
    child2 = parent2;
    
    child1.prog[random(0, progs_len-1)] = parent2.prog[random(0, progs_len-1)];
    child2.prog[random(0, progs_len-1)] = parent1.prog[random(0, progs_len-1)];
}

bool mutator(GA_chromosome &genome, UINT _pmut)
{
    bool mutated = false;

    if (random(0, unit) < _pmut)
    {
        mutated = true;
        UINT in = random(0, progs_len - 1); // kterou instrukci mutovat
        UINT part = random(0, unit);    // jakou cast instrukce mutovat
        
        if (part <= 33)   // muatace cele instrukce
        {
            genome.prog[in].opc = random(0, INSS - 1);
            set_args(genome.prog[in]);
        }
        else if (part <= 66)    // mutace arg1
        {
            // mutaci argumentu s ohledem na typ instrukce
            switch (genome.prog[in].opc)
            {
                case SET: case INC: case DEC:
                    genome.prog[in].arg1 = random(v0, v3); break;
                case REP: genome.prog[in].arg1 = random(0, p1); break;
                case GEN: genome.prog[in].arg1 = random(0, BLOCKS - 1); break;
                default: break;
            }
        }
        else    // mutace arg2
        {
            switch (genome.prog[in].opc)
            {
                case SET: genome.prog[in].arg2 = random(0, p1); break;
                case INC: case DEC: case REP:
                    genome.prog[in].arg2 = random(0, max_const); break;
                default: break;
            }
        }
    }

    return mutated;
}

ostream & operator<<(ostream &os, const Cell &c)
{
    switch (c.block)
    {
#ifdef _ID_V
        case ID_V: os << "ID_V       "; break;
#endif
#ifdef _ID_HV
        case ID_HV: os << "ID_HV      "; break;
#endif
#ifdef _AND
        case AND: os << "AND(" << c.in1 << ", " << c.in2 << ")  "; break;
#endif
#ifdef _HA_D
        case HA_D: os << "HA_D       "; break;
#endif
#ifdef _HA_AV
        case HA_AV: os << "HA_AV(" << c.in1 << ", " << c.in2 << ")"; break;
#endif
#ifdef _HA_AH
        case HA_AH: os << "HA_AH(" << c.in1 << ", " << c.in2 << ")"; break;
#endif
#ifdef _FA_D
        case FA_D: os << "FA_D       "; break;
#endif
#ifdef _FA_A
        case FA_A: os << "FA_A(" << c.in1 << ", " << c.in2 << ") "; break;
#endif
#ifdef _FA_AD
        case FA_AD: os << "FA_AD(" << c.in1 << ", " << c.in2 << ")"; break;
#endif
        default: break;
    }

    return os;
}

ostream & operator<<(ostream &os, const Cell circ[rows][cols])
{
    for (UINT i = 1; i < rows; i++)
    {
        bool level = false;
        for (UINT j = 1; j < cols; j++)
        {
            if (circuit[i][j].active)
            {
                level = true;
                if (j > 1) os << '\t';
                os << circuit[i][j];
            }
        }
        if (level) os << endl;
    }
    
    return os;
}

ostream & operator<<(ostream &os, const BLOCK &bl)
{
    switch (bl)
    {
#ifdef _ID_V
        case ID_V: os << "ID_V"; break;
#endif
#ifdef _ID_HV
        case ID_HV: os << "ID_HV"; break;
#endif
#ifdef _AND
        case AND: os << "AND"; break;
#endif
#ifdef _HA_D
        case HA_D: os << "HA_D"; break;
#endif
#ifdef _HA_AV
        case HA_AV: os << "HA_AV"; break;
#endif
#ifdef _HA_AH
        case HA_AH: os << "HA_AH"; break;
#endif
#ifdef _FA_D
        case FA_D: os << "FA_D"; break;
#endif
#ifdef _FA_A
        case FA_A: os << "FA_A"; break;
#endif
#ifdef _FA_AD
        case FA_AD: os << "FA_AD"; break;
#endif
        default: break;
    }

    return os;
}

ostream & operator<<(ostream &os, const Inst &i)
{
#ifdef DEBUG
    switch (i.opc)
    {
        case SET: os << "SET "; break;
        case INC: os << "INC "; break;
        case DEC: os << "DEC "; break;
        case REP: os << "REP "; break;
        case GEN: os << "GEN "; break;
        case NOP: os << "NOP "; break;
        default: break;
    }

    if (i.opc == NOP) return os;
    if (i.opc == GEN) { os << (BLOCK) i.arg1 << ' '; return os; }
    
    if (i.arg1 >= p0) os << "p" << i.arg1 - p0 << ' ';
    else if (i.arg1 >= v0) os << "v" << i.arg1 - v0 << ' ';
    else os << i.arg1 << '\t';
    
    if (i.arg2 >= p0) os << "p" << i.arg2 - p0 << ' ';
    else if (i.arg2 >= v0) os << "v" << i.arg2 - v0<<' ';
    else os << i.arg2;
#else
    os << i.opc << " " << i.arg1 << " " << i.arg2 << " ";
#endif

    return os;
    
}

ostream & operator<<(ostream &os, const GA_chromosome &g)
{
#ifdef DEBUG
    UINT most_long = 0;
    for (UINT i = 0; i < sizeof(aprog_len) / sizeof(aprog_len[0]); i++)
        if (aprog_len[i] > most_long) most_long = aprog_len[i];

    for (UINT i = 0; i < most_long; i++)
    {
        for (UINT j = 0; j < sizeof(aprog_len) / sizeof(aprog_len[0]); j++)
        {
            if (i < aprog_len[j]) os << g.prog[offset[env[j]] + i] << '\t';
            else os << "         " << '\t';
        }
        cout << endl;
    }
#else
	for (UINT i = 0; i < progs_len; i++)
        os << g.prog[i];
#endif

	return os;
}

double pop_fit_avg()
{
    double sum = 0.0;
    
    for (UINT i = 0; i < popsize; i++)
        sum += population[i].fitness;
        
    return (sum / (double) popsize);
}

bool stop()
{
    if (best.fitness > best_ever)
    {
#ifdef DEBUG
		cout << "Fitness = " << best.fitness << "/" << (UINT) max_fit
									<< " in generation " << generation << endl;
#endif
		best_ever = best.fitness;
    }
	
    // zaznam prubehu fitness	
    if (generation == 1 || generation % avgfcalc == 0 || best_ever == max_fit)
    {
        pop_fit[fit_ix].pop_avg = pop_fit_avg();
        pop_fit[fit_ix].best = best_ever;
        pop_fit[fit_ix].gen = generation;
        fit_ix++;
    }
    
    if (best_ever == max_fit || generation == generations)
	{
		ofstream ofs("results", ios_base::app);
		if (best_ever == max_fit)
        {
        	    ofs << "gen. " << generation << endl << best << endl;
                fit_ix = 0; do
                {
                    ofs << pop_fit[fit_ix].gen << " " << pop_fit[fit_ix].pop_avg
                        << " " << pop_fit[fit_ix].best << endl;
                    fit_ix++;
                } while (pop_fit[fit_ix - 1].best != max_fit);
        }
		else ofs << "FAILED" << endl;
		ofs.close();
		return true;
	}
    
    return false;
}

// evolucni cyklus: nova populace o popsize jedincich je tvorena krizenimi
// a mutaci turnajem vybranymi jedinci predchazejici generace
// specifikum: elitismus nahrazuje jedince s totoznymi fitness - vyrazny uspech!
void evolve()
{
	// inicializace
	generation = 0;
	best.fitness = 0;
	best_ever = 0.0;
    fit_ix = 0;
	GA_chromosome ind1_new, ind2_new;
	UINT _tour = tour >= 2 ? tour : 2;
	UINT i1;
    
	for (UINT i = 0; i < popsize; i++)
	{
		initialize(pool1[i]);
		pool1[i].evaluate = true;
	}
	// evolucni cyklus
	do
	{
		generation++;
		if (generation & 1)
		{
			population = pool1;
			next_population = pool2;
		}
		else
		{
			population = pool2;
			next_population = pool1;
		}
		// ohodnoceni populace
		for (UINT i = 0; i < popsize; i++)
		{
			if (population[i].evaluate)
			{
				population[i].fitness = fitness(population[i]);
				if (population[i].fitness >= best.fitness) best = population[i];
				population[i].evaluate = false;
			}
		}
		// elitizmus
		next_population[0] = best;
        GA_chromosome mutant = best; mutator(mutant, unit);
		next_population[1] = mutant;
		// tvorba nove populace
		for (UINT i = 2; i < popsize; i += 2)
		{
			GA_chromosome *ind1 = NULL, *ind2 = NULL;
			// turnajovy vyber jedincu
			for (UINT t = 0; t < _tour; t++)
			{
				i1 = random(0, popsize - 1);
				if (ind1 == NULL) ind1 = &population[i1];
				else if (ind2 == NULL) ind2 = &population[i1];
				else if (population[i1].fitness > ind1->fitness)
					ind1 = &population[i1];
				else if (population[i1].fitness > ind2->fitness)
					ind2 = &population[i1];
			}
			// krizeni
			if (random(0, unit) < pcross)
			{
				crossover(*ind1, *ind2, ind1_new, ind2_new);
				ind1_new.evaluate = true;
				ind2_new.evaluate = true;
			}
			else	// prechod jedincu bez krizeni
			{
				ind1_new = *ind1;
				ind2_new = *ind2;
			}
			// mutace
			if (mutator(ind1_new, pmut)) ind1_new.evaluate = true;
			if (mutator(ind2_new, pmut)) ind2_new.evaluate = true;
			// umisteni potomku do nove populace
			next_population[i] = ind1_new;
			next_population[i + 1] = ind2_new;
		}
	} while (!stop());
}

void system_init()
{
    // inicializace offsetu programu genomu
    offset[0] = 0;
    for (UINT i = 1; i < sizeof(aprog_len) / sizeof(aprog_len[0]); i++)
        offset[i] = offset[i - 1] + aprog_len[i - 1];
    
    // inicializace poctu testovacich vektoru
    for (UINT i = 2; i <= bitw; i++)
    {
		if (i < 6) sequences[i] = 32;
		else sequences[i] = (UINT) pow(2.0, (int) i);
		vectors[i] = (UINT) sequences[i] / (8 * sizeof(UINT));
    }
	max_fit = (double) sequences[bitw] * (double) bitw;
}

void help(char name[])
{
	   cout << "usage: " << name << " [FILENAME]" << endl;
	   cout << "If FILENAME is supplied, the system executes the evolved programs" << endl;
	   cout << "contained in the file, evaluates the multipliers developed uing these" << endl;
	   cout << "programs and writes the structures of solutions satisfying the operand" << endl;
	   cout << "widths specified in the application to the standard output." << endl;
	   cout << "If no file is specified, evolutionary process is started." << endl;
}

int main(int argc, char *argv[])
{
    system_init();

    if (argc > 1)
    {
        if (strncmp(argv[1], "--help", 65536) == 0)
        {
           help(argv[0]);
           return 0;
        }

        // test vyevolvovanych konstruktoru
        ifstream ifs(argv[1]);
        GA_chromosome g;
        UINT order = 1;
        ifs >> g.prog[0].opc >> g.prog[0].arg1 >> g.prog[0].arg2;
        while (!ifs.eof())
        {
            for (UINT i = 1; i < progs_len; i++)
                ifs >> g.prog[i].opc >> g.prog[i].arg1 >> g.prog[i].arg2;
            if (fitness(g) == max_fit)
            {
                cout << g << endl;
                cout << circuit << endl;
            }
            else cout << order << endl;
            
            ifs >> g.prog[0].opc >> g.prog[0].arg1 >> g.prog[0].arg2;
            order++;
        }
        ifs.close();
    }
    else
    {
        while (1)
        {
            FILE *rand = fopen("/dev/random", "r");
            UINT r;
            fread(&r, sizeof(r), 1, rand);
            fclose(rand);
            srand(r);
            evolve();
        }
    }
    
	return 0;
}
